//icaclient/develop/main/UKH/NetClient/main/linux/obj.h264/ctxh264.cpp#41 - edit change 297196 (ktext)
/*****************************************************************************
 *
 *   ctxh264.cpp
 *
 *   H.264 v3 interface software fallback implementation.
 *
 *   $Id: //icaclient/unix13.3/client/unix/CitrixPluginSDK/H264_sample/ctxh264.cpp#1 $
 *
 *   Copyright 2013-2015 Citrix Systems, Inc.  All Rights Reserved.
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include <execinfo.h>
#include <unistd.h>
#include <time.h>
#include <dlfcn.h>
#include <algorithm>
#include <sys/shm.h>
#include <sys/ipc.h>

#define X11_SUPPORT // FIXME!
#ifdef X11_SUPPORT
#include <X11/X.h>
#else
/* TODO non-X11 push code, fix interface to ff_decode_rgb().
 * Also fix mis-use of MAINLOOP interface.
 */
#endif /* X11_SUPPORT */

#include <linux/limits.h>

#include "citrix.h"
#include "UI_if.h"
#include "mainloop.h"

#ifndef PLUGIN_SDK
#include "hostfeat.h"
#endif

#include "FFDecode.h"

/* Define "Decoder" structure for use by ctxh264.h. */

struct Decoder {
    FFMPEGCtx *pFFCtx;
};

/* FFMpeg function forward decls. */

bool ff_decode_rgb(FFMPEGCtx* pCtx, PBYTE pInData, UINT32 uInSize,
		   XImage *img, UINT32 uWidth, UINT32 uHeight);
bool ff_decode_yuv(FFMPEGCtx *pCtx, PBYTE pInData, UINT32 uInSize,
		   char *data, int *offsets, int *pitches,
		   UINT32 uWidth, UINT32 uHeight);

#include "ctxh264.h"
#include "h264_common.cpp" // FIXME!

struct H264_decoder H264_decoder = {
    VERSION_MAJOR,
    VERSION_MINOR,
    32,
    4096,
    2048,
    60,
    H264_OPTION_LOSSLESS | H264_OPTION_PREFER_TEXT_RECTS |
    H264_OPTION_WINDOW_SUPPORT |     /* Seamless */
    H264_OPTION_SMALL_FRAME_SUPPORT,
    H264_CHROMA_FORMAT_444,
    0,                 /* Preferred alpha value for lossless objects. */
    PIXEL_FORMAT_ARGB, /* Preferred pixel format for lossless objects. */
    &v3_init,
    &v3_open_context,
    &v3_start_frame,
    &v3_decode_frame,
    &v3_compose_with_fb,
    &v3_compose_with_rects,
    &v3_push_frame,
    &v3_close_context,
    &v3_end,
};

/* This function would be called only once, to initialize the DLL. */

bool v3_init()
{
    char ExePath[PATH_MAX];

#ifdef X11_SUPPORT
    if (!X11_init())
        return false;
#endif

    /* Initialize custom ffmpeg decoding. */

    snprintf(ExePath, sizeof ExePath, "%s/lib", getICARoot());
    ff_init(ExePath);
    return true;
}

void close_context(V3FBCtxData *pCtxData)
{
    ff_close_ctx(pCtxData->dec.pFFCtx);
    close_context_common(pCtxData);
}

void v3_end ()
{
  end_common();
  ff_end();
}

H264_context v3_open_context(int width, int height, void *codec_data,
			     int len, unsigned int options)
{
    V3FBCtxData *pCtxData;

    pCtxData = open_context_common(width, height, codec_data, len, options);
    if (pCtxData) {
        pCtxData->dec.pFFCtx = ff_open_ctx();
        return pCtxData->CtxId;
    }
    return H264_INVALID_CONTEXT;
}

bool v3_start_frame(H264_context Ctx, unsigned int encoded_size,
                    SIGNED_RECT dirty_rects[], unsigned int num_rects)
{
    V3FBCtxData *pCtxData;

    /* FFMPEG decoder requires us to allocate 6 bytes extra. See
     * ff_decode() for more details.
     *
     * When start_frame_common() returns non-NULL, mutex is locked.
     */

    if (encoded_size > 0)
        encoded_size += H264_NAL_UNIT_DELIM_SIZE;
    pCtxData = start_frame_common(Ctx, encoded_size);
    if (pCtxData) {
        copy_dirty_region(pCtxData, dirty_rects, num_rects);
        pthread_mutex_unlock(&pCtxData->H264DataLock);
        return true;
    }
    return false;
}

bool v3_decode_frame(H264_context Ctx, void *H264_data, int len, bool last)
{
    V3FBCtxData *pCtxData;

    pCtxData = decode_frame_common(Ctx, H264_data, len, last);
    if (!pCtxData)
        return false;
    if (last) {
        if (pCtxData->fb_yuv) {
            ff_decode_yuv(pCtxData->dec.pFFCtx, pCtxData->pEncodedFrame, 
                    pCtxData->uEncFrameSize, pCtxData->fb_yuv->data,
                    pCtxData->fb_yuv->offsets, pCtxData->fb_yuv->pitches, 
                    pCtxData->uCtxWidth, pCtxData->uCtxHeight);
        } else {
            ff_decode_rgb(pCtxData->dec.pFFCtx, pCtxData->pEncodedFrame, 
                    pCtxData->uEncFrameSize, pCtxData->fb, 
                    pCtxData->uCtxWidth, pCtxData->uCtxHeight);
        }
        pCtxData->uEncFrameCurSize = 0;
        pCtxData->uEncFrameSize = 0;
    }
    pthread_mutex_unlock(&pCtxData->H264DataLock);
    return true;
}
